//Refresh time interval in msec
//IMPORTANT! Interval size must be more then 30 sec. Otherwise, the client will be blocked
//The script use three objects of Nest (Air, Cam, Thermostat) so that time is 90000
var UPDATEINTERVAL = 90000;

//You should get CLIENT_ID and CLIENT_SECRET by registering you app at https://developers.nest.com/
var CLIENT_ID = "";
var CLIENT_SECRET = "";
/**
 * The listener contains mandatory elemets which need initialization
 */
IR.AddListener(IR.EVENT_START, 0, function() {
    /*
    * Array of objects for set images in gui elements
    * it is parametr of Nest camera object
    * Zero element is first camera(Camera 1)
    * Page is name of page on which the element is located 
    * Item is name of element on which the image is located
    * Click is name of element which by press will be open browser or app 
    * Public URL need for displaying public url if it is not empty 
    * RunInAppInsteadBrowser mobile devices can run video in Nest app    
        Getting Publick URL
        1) Go to your nest accaunt
        2) Select your camera
        3) Select settungs in bottom corner 
        4) List to and select "Camera sharing"
        5) Select "Share publicly" 
        6) Select "Agree & Share"
        7) Take a share link
        8) Insert it in your iRidium app
    */
    var l_aCamsItems = [{
            Page: "Main",
            Item: "Image",
            Click: "Image",
            PublickURL: "",
            RunInAppInsteadBrowser: false
        }]
        /* 
         * For working all devices Nest need a one central device, 
         * for sending request, run authorization and monitoring condition
         * Channels:
         * -> Authorize //Run browser for authorization
         * -> Logout //Erase tokens and set state authorization is false
         * Tags:
         * <- Authorized //Contains state of authorization
         * <- Error //Contains last error API nest
         */
    var CentralNestDevice = IR.GetDevice("CentralNestDevice");

    /*
     * Object of Nest queue
     * for sending commands one by one
     * input parameter is central device
     */
    var MyNestQueue = new NestQueue(CentralNestDevice);

    /*
     * Object for Nest OAuth 2.0 authorization
     * monitors the channels and start authorization 
     * input parametr is central device
     */
    var MyAuthNest = new AuthNest(CentralNestDevice);

    /*
     * Object of Nest cameras
     * Gets information about all cameras and set it according to the tags of its driver
     */
    var MyCamNest = new DriverCamNest(IR.GetDevice("NestCams"), CentralNestDevice, MyNestQueue, l_aCamsItems);
    /*
     * Object of Nest Air detectors
     * Gets information about all air detectors and set it according to the tags of its driver
     */
    var MyAirNest = new DriverNestAridDetectors(IR.GetDevice("NestAridDetectors"), CentralNestDevice, MyNestQueue);

    /*
     * Object of Nest Termostats
     * Gets information about all Termostats, set it according to the tags of its driver
     * and monitors the channels and send commands
     */
    var MyTermostatNest = new DriverTermostats(IR.GetDevice("NestTermostats"), CentralNestDevice, MyNestQueue);

    /*
     * It is enough to initialize a poll of one of the drivers
     * The rest of the drivers automatically start, after the first query is executed, 
     * they are initialized by changing the tag of authorization state
     */
    MyCamNest.Start();
    MyAirNest.Start();
    MyTermostatNest.Start();

});
/**
 * Constructor of OAuth 2.0 NEST authorization object
 * The object saves the key in to Global.AccesNestToken
 * Start authorization initializing by external method Start(in_fCallBack) or by set authorization channel of central device
 * Input parametr is central device for set authorization state and listens channels
 * @constructor
 */
var AuthNest = function(in_oDriver) {
    var CentralNestDevice = in_oDriver;
    var DeviceNest = IR.CreateDevice(IR.DEVICE_CUSTOM_HTTP_TCP, "NestToken", "api.home.nest.com", 443, "", "", true);
    DeviceNest.SetParameters({
        ScriptMode: 1
    });
    DeviceNest.Connect();

    var out_fCallBack = function() {};

    /*
     * Function of extracting auth code from uri 
     * for preparing next request
     * Your uri may be is lochalhost
     */
    function getCode(in_sUrl) {
        if (in_sUrl.indexOf("https://iridiummobile.com/?state=STATE&code=") === 0) {
            var l_sCode = in_sUrl.substr(44);
            IR.CloseBrowser();
            DeviceNest.SendEx({
                Type: "POST",
                Url: '/oauth2/access_token',
                Data: ['client_id=' + CLIENT_ID + "&client_secret=" + CLIENT_SECRET + "&code=" + l_sCode + "&grant_type=authorization_code"],
                cbReceiveText: function(in_sText) {
                    CentralNestDevice.SetFeedback("Authorized", true);
                    try {
                        var l_oData = JSON.Parse(in_sText);
                    } catch (err) {
                        IR.Log("Error of receive data.")
                    }

                    if (l_oData instanceof Object) {
                        if (l_oData.access_token) {
                            IR.SetVariable("Global.AccesNestToken", l_oData.access_token);
                            out_fCallBack();
                        }

                        DeviceNest.Disconnect();
                    }
                    removeBrowserListeners();
                }
            });
        }
    }

    /*
     * Remove listeners to avoid bouble triggering
     */
    function removeBrowserListeners() {
        IR.RemoveListener(IR.EVENT_BROWSER_PAGE_FINISHED, 0, getCode);
        IR.RemoveListener(IR.EVENT_BROWSER_PAGE_ERROR, 0, getCode);
    }

    /**
     * Start authorization. Parametr is link of callback function after successful authorization
     * @param in_fCallback {function}
     */
    this.Start = function(in_fCallback) {
        if (in_fCallback)
            out_fCallBack = in_fCallback;
        IR.OpenBrowser("https://home.nest.com/login/oauth2?client_id=" + CLIENT_ID + "&state=STATE");
        IR.AddListener(IR.EVENT_BROWSER_PAGE_FINISHED, 0, getCode);
        IR.AddListener(IR.EVENT_BROWSER_PAGE_ERROR, 0, getCode);
    };

    IR.AddListener(IR.EVENT_CHANNEL_SET, CentralNestDevice, function(in_sChannel, in_sValue) {
        if (in_sChannel === "Authorize")
            this.Start();
        if (in_sChannel === "Logout") {
            IR.SetVariable("Global.AccesNestToken", "");
            CentralNestDevice.SetFeedback("Authorized", false);
        }
    }, this);

};
/*
 * Constructor of queue Nest
 * for sending commands one by one
 * Uses Bearer authorization with AccessToken Nest
 * The queue have external method Send
 * Input parametr is central device for working with it
 * @constructor
 */
var NestQueue = function(in_oDriver) {
    //Time for repeat send command after timeout
    var TIMOUTINTERVAL = 20000;
    //Count of repeats after timout
    var REPEATLIMIT = 2;

    //Array of queue
    var l_aQueue = [];

    //The current processing object
    var l_oProcessingQueueElement = {};

    //Counter of repeats current processing object after timout
    var l_nNowRepeats = 0;
    //State of waiting for current response
    var l_bWaitResponse = false;
    //Object of timeout
    var l_oTimout = {};

    var DeviceNest = in_oDriver;
    /**
     * External method for sending requests
     * @in_sType {string} - Type (GET or POST)
     * @in_sUrl {string} - Uri without api(api.nest.com) uri. Example "/devices/cameras"
     * @in_fCallBack {function} - Callback function for response
     * @in_oData {object} - Data, if request is POST
     */
    this.send = function(in_sType, in_sUrl, in_fCallBack, in_oData) {

        var l_oAddQueueElement = {};

        l_oAddQueueElement.Type = in_sType;
        l_oAddQueueElement.Url = in_sUrl;
        l_oAddQueueElement.fCallBack = in_fCallBack;

        if (in_oData)
            l_oAddQueueElement.Data = in_oData;
        else
            l_oAddQueueElement.Data = [];


        IR.Log(l_oAddQueueElement.Url)

        l_aQueue.push(l_oAddQueueElement);

        if (l_bWaitResponse === false) {
            SendNext();
        }

    };

    //Internal method for send current request
    function QueueSend(in_oQueueElement) {
        Wait(true);
        DeviceNest.SendEx({
            Type: in_oQueueElement.Type,
            Url: in_oQueueElement.Url,
            Data: [in_oQueueElement.Data],
            Headers: {
                "Content-Type": "application/json",
                "Authorization": "Bearer " + IR.GetVariable("Global.AccesNestToken")
            },
            cbReceiveText: function(in_sText, in_sCode) {
                Wait(false);
                if (in_sText.indexOf("error") === -1) {
                    in_oDriver.SetFeedback("Authorized", true);
                    in_oQueueElement.fCallBack(in_sText);
                    if (l_aQueue.length > 0)
                        SendNext();
                } else {
                    var l_oError = JSON.Parse(in_sText);
                    DeviceNest.SetFeedback("Error", l_oError.message);
                    if (l_aQueue.length > 0)
                        SendNext();
                    if (in_sText.indexOf("unauthorized") !== -1)
                        in_oDriver.SetFeedback("Authorized", false);
                }
            }
        });
    }

    /*
     * Internal method for set current state of waiting
     * На входе @in_bState {boolean}
     */
    function Wait(in_bState) {
        if (in_bState === true) {
            l_bWaitResponse = true;
            IR.ClearInterval(l_oTimout);
            l_oTimout = IR.SetTimeout(TIMOUTINTERVAL, Timout);
        } else {
            l_bWaitResponse = false;
            IR.ClearInterval(l_oTimout);
        }
    }

    /*
     * Internal method for sending next element of queue
     */
    function SendNext() {
        l_oProcessingQueueElement = l_aQueue.shift();
        QueueSend(l_oProcessingQueueElement);
    }

    /*
     * Internal method executing when is timeout is over
     * Running repeat untill repeats is less REPEATLIMIT
     * Otherwise, it deletes the current element of the queue, or if there are other elements
     * starts sending the next one
     */
    function Timout() {
        if (l_nNowRepeats < REPEATLIMIT) {
            l_nNowRepeats++;
            QueueSend(l_oProcessingQueueElement);
        } else {
            l_nNowRepeats = 0;
            if (l_aQueue.length > 0)
                SendNext();
        }
    }
};


/**
 * Object of Nest termostats
 * In input:
 * device of cams @in_oCamsDriver
 * central device @in_oDriver
 * and queue @in_oQueue
 * @constructor
 */
var DriverTermostats = function(in_oCamsDriver, in_oDriver, in_oQueue) {
    var MyNestQueue = in_oQueue;
    var RefreshInterval;
    var l_bIsWork = false;

    /*
     * Listener for starting and stop interval depending by authorization state
     */
    IR.AddListener(IR.EVENT_TAG_CHANGE, in_oDriver, function(in_sChannel, in_sValue) {
        if (in_sChannel === "Authorized")
            if (in_sValue === true) {
                if (!l_bIsWork)
                    this.Start();
            } else {
                l_bIsWork = false;
                IR.ClearInterval(RefreshInterval);
            }
    }, this);

    /*
     * Listener fo receiving channel states
     * for prepare and sending commands
     */
    IR.AddListener(IR.EVENT_CHANNEL_SET, in_oCamsDriver, function(in_sChannel, in_sValue) {
        var l_sDeviceID = in_oCamsDriver.GetFeedback(in_sChannel.split(":")[0] + ":device_id");
        var l_oPutData = {};
        switch (in_sChannel.split(":")[1]) {
            case "hvac_mode_cool":
                in_oCamsDriver.SetFeedback(in_sChannel.split(":")[0] + ":hvac_mode", "cool");
                l_oPutData["hvac_mode"] = "cool";
                MyNestQueue.send("PUT", "/devices/thermostats/" + l_sDeviceID, CBChannelSet, JSON.Stringify(l_oPutData));
                break;
            case "hvac_mode_heat":
                in_oCamsDriver.SetFeedback(in_sChannel.split(":")[0] + ":hvac_mode", "heat");
                l_oPutData["hvac_mode"] = "heat";
                MyNestQueue.send("PUT", "/devices/thermostats/" + l_sDeviceID, CBChannelSet, JSON.Stringify(l_oPutData));
                break;
            case "hvac_mode_eco":
                in_oCamsDriver.SetFeedback(in_sChannel.split(":")[0] + ":hvac_mode", "eco");
                l_oPutData["hvac_mode"] = "eco";
                MyNestQueue.send("PUT", "/devices/thermostats/" + l_sDeviceID, CBChannelSet, JSON.Stringify(l_oPutData));
                break;
            case "hvac_mode_heatcool":
                in_oCamsDriver.SetFeedback(in_sChannel.split(":")[0] + ":hvac_mode", "heat-cool");
                l_oPutData["hvac_mode"] = "heat-cool";
                MyNestQueue.send("PUT", "/devices/thermostats/" + l_sDeviceID, CBChannelSet, JSON.Stringify(l_oPutData));
                break;
            case "hvac_mode_off":
                in_oCamsDriver.SetFeedback(in_sChannel.split(":")[0] + ":hvac_mode", "off");
                l_oPutData["hvac_mode"] = "off";
                MyNestQueue.send("PUT", "/devices/thermostats/" + l_sDeviceID, CBChannelSet, JSON.Stringify(l_oPutData));
                break;
            case "fan_timer_active":
            case "temperature_scale":
            case "target_temperature_f":
            case "target_temperature_c":
            case "target_temperature_high_f":
            case "target_temperature_high_c":
            case "target_temperature_low_f":
            case "target_temperature_low_c":
            case "hvac_mode":
            case "label":
            case "fan_timer_duration":
                in_oCamsDriver.SetFeedback(in_sChannel.split(":")[0] + ":" + in_sChannel.split(":")[1], in_sValue);
                l_oPutData[in_sChannel.split(":")[1]] = in_sValue;
                MyNestQueue.send("PUT", "/devices/thermostats/" + l_sDeviceID, CBChannelSet, JSON.Stringify(l_oPutData));
                break;
        }
    });

    /*
     * Empty function
     */
    function CBChannelSet(in_sText) {
        //IR.Log(in_sText);
    }

    /*
     * Function for sending a request for devices states
     */
    function refreshTermostatList() {
        MyNestQueue.send("GET", "/devices/thermostats", CBTermostatList);
    }

    /*
     * The function of processing the answer of the list of elements with their parameters
     */
    function CBTermostatList(in_sText) {
        var l_oTermostats = {};
        try {
            l_oTermostats = JSON.Parse(in_sText);
        } catch (err) {
            IR.Log("Error receive Termostat data");
        }
        if (l_oTermostats instanceof Object) {
            var l_nCount = 1;
            for (var key in l_oTermostats) {
                for (var in_key in l_oTermostats[key]) {
                    in_oCamsDriver.SetFeedback("Termostat " + l_nCount + ":" + in_key, l_oTermostats[key][in_key]);
                }
                l_nCount++;
            }
        }
    }

    /*
     * External method for start interval
     */
    this.Start = function() {
        l_bIsWork = true;
        refreshTermostatList();
        IR.ClearInterval(RefreshInterval);
        RefreshInterval = IR.SetInterval(UPDATEINTERVAL, refreshTermostatList)
    };
};
/**
* Object of Nest cameras
* In input:
* Device of cameras @in_oCamsDriver
* Central device@in_oDriver
* Till absent possibility using tag for image value need add @in_aCamsItems
    * Array of objects for set images in gui elements
    * it is parametr of Nest camera object
    * Zero element is first camera(Camera 1)
    * Page is name of page on which the element is located 
    * Item is name of element on which the image is located
    * Click is name of element which by press will be open browser or app 
    * Public URL need for displaying public url if it is not empty 
    * RunInAppInsteadBrowser mobile devices can run video in Nest app    
        Getting Publick URL
        1) Go to your nest accaunt
        2) Select your camera
        3) Select settungs in bottom corner 
        4) List to and select "Camera sharing"
        5) Select "Share publicly" 
        6) Select "Agree & Share"
        7) Take a share link
        8) Insert it in your iRidium app
    var l_aCamsItems = 
    [
        {Page: "Main", Item: "Image", Click: "Image", PublickURL: "", RunInAppInsteadBrowser: false} 
    ]
* @constructor
*/
var DriverCamNest = function(in_oCamsDriver, in_oDriver, in_oQueue, in_aCamsItems) {
    var MyNestQueue = in_oQueue;
    var RefreshInterval;
    var l_bIsWork = false;

    var l_aCamsItems = in_aCamsItems;

    /*
     * Listener for starting and stop interval depending by authorization state
     */
    IR.AddListener(IR.EVENT_TAG_CHANGE, in_oDriver, function(in_sChannel, in_sValue) {
        if (in_sChannel === "Authorized")
            if (in_sValue === true) {
                if (!l_bIsWork)
                    this.Start();
            } else {
                l_bIsWork = false;
                IR.ClearInterval(RefreshInterval);
            }
    }, this);

    /*
     * Adding listeners for oppening browser or app
     */
    for (var i = 0; i < l_aCamsItems.length; i++) {
        if (IR.GetPage(l_aCamsItems[i].Page)) {
            IR.AddListener(IR.EVENT_ITEM_RELEASE, IR.GetPage(l_aCamsItems[i].Page).GetItem(l_aCamsItems[i].Click), function() {
                if (IR.GetVariable("System.OS.Name") == "Apple iOS" || IR.GetVariable("System.OS.Name") == "Android") {
                    if (l_aCamsItems[this.itemNum].RunInAppInsteadBrowser === true)
                        IR.Execute(in_oCamsDriver.GetFeedback("Camera " + (this.itemNum + 1) + ":app_url"));
                    else
                    if (l_aCamsItems[this.itemNum].PublickURL === "")
                        IR.Execute(in_oCamsDriver.GetFeedback("Camera " + (this.itemNum + 1) + ":web_url")); // web_url app_url
                    else
                        IR.Execute(l_aCamsItems[this.itemNum].PublickURL.replace("live", "embedded/live"));
                } else {
                    if (l_aCamsItems[this.itemNum].PublickURL === "")
                        IR.Execute(in_oCamsDriver.GetFeedback("Camera " + (this.itemNum + 1) + ":web_url"));
                    else
                        IR.Execute(l_aCamsItems[this.itemNum].PublickURL.replace("live", "embedded/live"));
                }
            }, {
                itemNum: i
            });
        } else {
            IR.Log("Page with name '" + l_aCamsItems[i].Page + "' is missing. Change name in l_aCamsItems.");
        }

    }

    /*
     * When image tag will be changed, listener can send tag value to image value of elements
     */
    IR.AddListener(IR.EVENT_TAG_CHANGE, in_oCamsDriver, function(in_sChannel, in_sValue) {
        if (in_sChannel.split(":")[1] === "snapshot_url") {
            var l_nElemNumber = in_sChannel.split(":")[0].split(" ")[1] - 1;
            if (l_nElemNumber < l_aCamsItems.length)
                IR.GetPage(l_aCamsItems[l_nElemNumber].Page).GetItem(l_aCamsItems[l_nElemNumber].Item).GetState(0).Image = in_sValue;
        }
    });

    /*
     * Function for sending a request for devices states
     */
    function refreshCameraList() {
        MyNestQueue.send("GET", "/devices/cameras", CBCameraList);
    }

    /*
     * The function of processing the answer of the list of elements with their parameters
     */
    function CBCameraList(in_sText) {
        var l_oCams = {};
        try {
            l_oCams = JSON.Parse(in_sText);
        } catch (err) {
            IR.Log("Error receive Camera data");
        }
        if (l_oCams instanceof Object) {
            var l_nCount = 1;
            for (var key in l_oCams) {
                for (var in_key in l_oCams[key]) {
                    if (in_key === "last_event") {
                        for (var l_sEvent in l_oCams[key][in_key]) {
                            in_oCamsDriver.SetFeedback("Camera " + l_nCount + ":last_event_" + in_key, l_oCams[key][in_key][l_sEvent]);
                        }
                    } else {
                        in_oCamsDriver.SetFeedback("Camera " + l_nCount + ":" + in_key, l_oCams[key][in_key]);
                    }
                }
                l_nCount++;
            }
        }
    }

    /*
     * External method for start interval
     */
    this.Start = function() {
        l_bIsWork = true;
        refreshCameraList();
        IR.ClearInterval(RefreshInterval);
        RefreshInterval = IR.SetInterval(UPDATEINTERVAL, refreshCameraList)
    };
};

/**
 * Object of Nest Air Detectors
 * In input:
 * device of cams @in_oCamsDriver
 * central device @in_oDriver
 * and queue @in_oQueue
 * @constructor
 */
var DriverNestAridDetectors = function(in_oAirsDriver, in_oDriver, in_oQueue) {
    var MyNestQueue = in_oQueue;
    var RefreshInterval;
    var l_bIsWork = false;

    /*
     * Listener for starting and stop interval depending by authorization state
     */
    IR.AddListener(IR.EVENT_TAG_CHANGE, in_oDriver, function(in_sChannel, in_sValue) {
        if (in_sChannel === "Authorized")
            if (in_sValue === true) {
                if (!l_bIsWork)
                    this.Start();
            } else {
                l_bIsWork = false;
                IR.ClearInterval(RefreshInterval);
            }
    }, this);

    /*
     * Function for sending a request for devices states
     */
    function refreshAirsList() {
        MyNestQueue.send("GET", "/devices/smoke_co_alarms", CBAirList);
    }

    /*
     * The function of processing the answer of the list of elements with their parameters
     */
    function CBAirList(in_sText) {
        var l_oAirs = {};
        try {
            l_oAirs = JSON.Parse(in_sText);
        } catch (err) {
            IR.Log("Error receive Air data");
        }
        if (l_oAirs instanceof Object) {
            var l_nCount = 1;
            for (var key in l_oAirs) {
                for (var in_key in l_oAirs[key]) {
                    in_oAirsDriver.SetFeedback("Air " + l_nCount + ":" + in_key, l_oAirs[key][in_key]);
                }
                l_nCount++;
            }
        }
    }

    /*
     * External method for start interval
     */
    this.Start = function() {
        l_bIsWork = true;
        refreshAirsList();
        IR.ClearInterval(RefreshInterval);
        RefreshInterval = IR.SetInterval(UPDATEINTERVAL, refreshAirsList)
    };
};